import { WASI } from "@wasmer/wasi";
import { WasmFs } from "@wasmer/wasmfs";
import { RubyVM } from "./index";
export const DefaultRubyVM = async (rubyModule, options = { consolePrint: true }) => {
    const wasmFs = new WasmFs();
    const wasi = new WASI({
        bindings: Object.assign(Object.assign({}, WASI.defaultBindings), { fs: wasmFs.fs }),
    });
    if (options.consolePrint) {
        const originalWriteSync = wasmFs.fs.writeSync.bind(wasmFs.fs);
        wasmFs.fs.writeSync = function () {
            let fd = arguments[0];
            let text;
            if (arguments.length === 4) {
                text = arguments[1];
            }
            else {
                let buffer = arguments[1];
                text = new TextDecoder("utf-8").decode(buffer);
            }
            const handlers = {
                1: (line) => console.log(line),
                2: (line) => console.warn(line),
            };
            if (handlers[fd])
                handlers[fd](text);
            return originalWriteSync(...arguments);
        };
    }
    const vm = new RubyVM();
    const imports = {
        wasi_snapshot_preview1: wasi.wasiImport,
    };
    vm.addToImports(imports);
    const instance = await WebAssembly.instantiate(rubyModule, imports);
    await vm.setInstance(instance);
    wasi.setMemory(instance.exports.memory);
    // Manually call `_initialize`, which is a part of reactor modoel ABI,
    // because the WASI polyfill doesn't support it yet.
    instance.exports._initialize();
    vm.initialize();
    return {
        vm,
        wasi,
        fs: wasmFs,
        instance,
    };
};
